import flet as ft
from calet.calet_errors import *
from calet.calet_theme import *
from calet.calet_button import *

# app title bar (ok)
class ClAppBar(ft.UserControl):
    """Represents an app title bar to be used in Flet apps.
    """
    def __init__(self, theme:ClTheme, title:str, win_actions:list[ClWinButton|ClIconButton],
                 left_title:bool=False, title_icon:str=None, left_icon:str=None, 
                 content_size:int=16, bar_size:int=40, defined_sections:bool=False, scrollable_sections:str=None, 
                 expand:bool|int=False, transparent:bool=False, with_blur:bool=False, 
                 left_actions:list[ClTextButton|ClIconButton|ClModeButton|ClSwitch|ClMenuButton]=[],
                 right_actions:list[ClTextButton|ClIconButton|ClModeButton|ClSwitch]=[]):
        """Use this properties to personalize the app bar:\n
        ---
        - theme: is an instance of ```calet_theme.ClTheme``` with the colors set to paint the bar.
        - title: is the text to display as title in the app bar.
        - win_actions: is a list of ```calet_button.ClWinButton``` or ```calet_button.ClIconButton``` objects that will be placed in the window buttons zone of the app bar.
        - left_title: is a flag saying if the title must be displayed on the left or middle of the app bar. If left_actions were given the title will always be on the middle.
        - title_icon: is an icon to display on the left of the title text if ```left_title``` is False, else on the right. Can be a Flet.icons constant or a path to an image.
        - left_icon: is an icon to display in the left side of the app bar.
        - bar_size: is th custom size of the bar. If ```expand``` is not False this property will be ignored.
        - content_size: is a personalized size for the title, left icon and title icon.
        - defined_sections: is a flag saying if the app bar sections must be displayed as a visual defined block or not.
        - scrollable_sections: is the scroll mode for app bar sections. Can be 'left' for only left section scrollable_sections, 'right' for only right section scrollable or 'both' for both sections scrollable.
        - expand: is the responsive expansion of the app bar in his container. See ```expand``` Flet property for more information.
        - transparent: is a flag saying if the app bar must be displayed transparent or colored.
        - with_blur: is a flag saying if the app bar must be displayed with blur effect or not.
        - left_actions: is a list of ```calet_button.ClTextButton```, ```calet_button.ClIconButton```, ```calet_button.ClModeButton```, ```calet_button.ClSwitch``` or ```calet_button.ClMenuButton``` objects to be displayed in the left side of the app bar.
        - right_actions: is a list of ```calet_button.ClTextButton```, ```calet_button.ClIconButton```, ```calet_button.ClModeButton```, ```calet_button.ClSwitch``` objects to be displayed in the right side of the app bar.
        """
        # VALIDATION
        if not isinstance(theme, ClTheme):
            raise ClError(
                error="Argument Error: <<theme>> must be an instance of 'calet_theme.ClTheme' class."
            )
        if not isinstance(title, str):
            raise ClError(
                error="Argument Error: <<title>> must be string."
            )
        if not isinstance(left_title, bool):
            raise ClError(
                error="Argument Error: <<left_title>> must be boolean."
            )
        if title_icon is not None and not isinstance(title_icon, str):
            raise ClError(
                error="Argument Error: <<title_icon>> must be string."
            )
        if left_icon is not None and not isinstance(left_icon, str):
            raise ClError(
                error="Argument Error: <<left_icon>> must be string."
            )
        elif left_icon is not None and "/" in left_icon:
            file = None
            try:
                file = open(left_icon)
            except Exception:
                raise ClError(
                    error="Argument Error: <<left_icon>> is an invalid path or the file doesn't exist."
                )
            finally:
                file.close()
            if not left_icon.endswith(("png","jpeg")):
                raise ClError(
                    error="Argument Error: <<left_icon>> is an invalid file. Must be in PNG or JPEG format."
                )
        if not isinstance(content_size, int):
            raise ClError(
                error="Argument Error: <<content_size>> must be integer."
            )
        if not isinstance(bar_size, int):
            raise ClError(
                error="Argument Error: <<bar_size>> must be integer."
            )
        if not isinstance(defined_sections, bool):
            raise ClError(
                error="Argument Error: <<defined_sections>> must be boolean."
            )
        if scrollable_sections is not None and not isinstance(scrollable_sections, str):
            raise ClError(
                error="Argument Error: <<scrollable_sections>> must be string."
            )
        elif scrollable_sections is not None and scrollable_sections not in ("left","right","both"):
            raise ClError(
                error="Argument Error: <<scrollable_sections>> must be 'left', 'right' or 'both'."
            )
        if not isinstance(expand, (bool,int)):
            raise ClError(
                error="Argument Error: <<expand>> must be boolean or integer."
            )
        if not isinstance(transparent, bool):
            raise ClError(
                error="Argument Error: <<transparent>> must be boolean."
            )
        if not isinstance(with_blur, bool):
            raise ClError(
                error="Argument Error: <<with_blur>> must be boolean."
            )
        if not isinstance(win_actions, list):
            raise ClError(
                error="Argument Error: <<win_actions>> must be a list."
            )
        else:
            for i in range(len(win_actions)):
                if not isinstance(win_actions[i], (ClWinButton, ClIconButton)):
                    raise ClError(
                        error=f"Argument Error: <<win_actions[{i}]>> must be an instance of 'calet_button.ClWinButton' or 'calet_button.ClIconButton' class."
                    )
        for i in range(len(left_actions)):
            if not isinstance(left_actions[i], (ClTextButton, ClIconButton, ClModeButton, ClSwitch, ClMenuButton)):
                raise ClError(
                    error=f"Argument Error: <<left_actions[{i}]>> must be an instance of 'calet_button.ClTextButton', 'calet_button.ClIconButton','calet_button.ClModeButton', 'calet_button.ClSwitch' or 'calet_button.ClMenuButton' class."
                )
        for i in range(len(right_actions)):
            if not isinstance(right_actions[i], (ClTextButton, ClIconButton, ClModeButton, ClSwitch)):
                raise ClError(
                    error=f"Argument Error: <<right_actions[{i}]>> must be an instance of 'calet_button.ClTextButton', 'calet_button.ClIconButton','calet_button.ClModeButton' or 'calet_button.ClSwitch' class."
                )
        # INITIALIZATION
        super().__init__()
        self.theme = theme
        self.title = title
        self.win_actions = win_actions
        self.left_title = left_title
        self.title_icon = title_icon
        self.left_icon = left_icon
        self.content_size = content_size
        self.bar_size = bar_size
        self.defined_sections = defined_sections
        self.scrollable_sections = scrollable_sections
        self.expand = expand
        self.transparent = transparent
        self.with_blur = with_blur
        self.left_actions = left_actions
        self.right_actions = right_actions
    
    def build(self):

        # APP BAR CONTENT
        # - app icon
        if self.left_icon is not None:
            self.app_icon = ft.Icon(
                name=self.left_icon,
                color=self.theme.primary,
                size=self.content_size+4
            ) if "/" not in self.left_icon else ft.Image(
                src=self.left_icon,
                fit=ft.ImageFit.CONTAIN,
                width=self.content_size+4,
                height=self.content_size+4
            )
        # - app title
        self.app_title = ft.Text(
            value=self.title,
            color=self.theme.font_one,
            size=self.content_size,
            text_align=ft.TextAlign.LEFT if self.left_title else ft.TextAlign.CENTER
        )
        # - app title icon
        if self.title_icon is not None:
            self.app_title_icon = ft.Icon(
                name=self.title_icon,
                color=self.theme.font_one,
                size=self.content_size
            )
        # - left items
        self.left_items = ft.Row(
            expand=None if self.left_title and not self.left_actions else 1,
            spacing=5,
            alignment=ft.MainAxisAlignment.START,
            controls=[]
        )
        # -- adding app icon
        if self.left_icon is not None:
            self.left_items.controls.append(
                ft.Container(
                    alignment=ft.alignment.center,
                    content=self.app_icon
                )
            )
        # -- adding title in the left
        if self.left_title and not self.left_actions:
            # --> app title
            self.left_items.controls.append(
                ft.Container(
                    alignment=ft.alignment.center,
                    content=self.app_title
                )
            )
            # --> app title icon
            if self.title_icon is not None:
                self.left_items.controls.append(
                    ft.Container(
                        alignment=ft.alignment.center,
                        content=self.app_title_icon
                    )
                )
        # -- adding left actions and title in the middle
        else:
            # --> left actions
            self.left_items.controls.append(
                ft.Container(
                    expand=True if self.scrollable_sections in ("left","both") else None,
                    alignment=ft.alignment.center_left,
                    bgcolor=self.theme.transparent_05 if self.defined_sections else None,
                    padding=2 if self.defined_sections else None,
                    border_radius=5 if self.defined_sections else None,
                    content=ft.Row(
                        spacing=5,
                        scroll=ft.ScrollMode.ADAPTIVE,
                        controls=self.left_actions
                    )
                )
            )
            # - middle items
            self.mid_items = ft.Row(
                expand=1,
                spacing=5,
                alignment=ft.MainAxisAlignment.CENTER,
                controls=[]
            )
            # -- app title icon
            if self.title_icon is not None:
                self.mid_items.controls.append(
                    ft.Container(
                        alignment=ft.alignment.center,
                        content=self.app_title_icon
                    )
                )
            # -- app title
            self.mid_items.controls.append(
                ft.Container(
                    alignment=ft.alignment.center,
                    content=self.app_title
                )
            )
        # - right items
        self.right_items = ft.Row(
            expand=True if self.left_title and not self.left_actions else 1,
            spacing=5,
            alignment=ft.MainAxisAlignment.END,
            controls=[
                # -- right actions
                ft.Container(
                    expand=True if self.scrollable_sections in ("right","both") else None,
                    alignment=ft.alignment.center_right,
                    bgcolor=self.theme.transparent_05 if self.defined_sections else None,
                    padding=2 if self.defined_sections else None,
                    border_radius=5 if self.defined_sections else None,
                    content=ft.Row(
                        spacing=5,
                        scroll=ft.ScrollMode.ADAPTIVE,
                        controls=self.right_actions
                    )
                ),
                # -- win actions
                ft.Row(
                    spacing=5,
                    controls=self.win_actions
                )
            ]
        )

        # APP BAR
        return ft.WindowDragArea(
            content=ft.Container(
                bgcolor=self.theme.transparent if self.transparent else self.theme.background_one,
                blur=5 if self.with_blur else None,
                height=self.bar_size if not self.expand else None,
                padding=ft.padding.only(left=5, top=4, right=5, bottom=4),
                content=ft.Row(
                    spacing=0,
                    alignment=ft.MainAxisAlignment.START,
                    controls=[
                        self.left_items,
                        self.right_items
                    ] if self.left_title and not self.left_actions else [
                        self.left_items,
                        self.mid_items,
                        self.right_items
                    ]
                )
            )
        )
    
    def upd(self, title:str=None):

        if title is not None:
            self.title = title
            self.app_title.value = title
        self.update()

# app menu bar section (ok)
class ClMenuSection(ft.UserControl):
    """Represents a section of an app submenu bar to be used in ```calet_bar.ClSubmenuBar``` objects."""
    def __init__(self, theme:ClTheme, actions:list[list], lateral:bool=False, defined:bool=False, expand:bool|int=False):
        """Use this properties to personalize the menu section:\n
        ---
        - theme: is an instance of ```calet_theme.ClTheme``` with the colors set to paint the bar.
        - actions: is a list of lists of ```calet_button.ClTextButton```, ```calet_button.ClButton```, ```calet_button.ClAcceptButton```, ```calet_button.ClCancelButton```, ```calet_button.ClSelectButton```, ```calet_button.ClModeButton```, ```calet_button.ClIconButton```, ```calet_button.ClMenuButton```, ```calet_button.ClCheck```, ```calet_button.ClRadio```, ```calet_button.ClSwitch``` where each sublist will represent a different column of actions in the section.
        - defined: is a flag saying if the section must be displayed as a visual defined block or not.
        - lateral: is a flag saying if the section will be displayed in a normal or lateral menu.
        - expand: is the responsive expansion of the menu section in his container. See ```expand``` Flet property for more information.
        """
        # VALIDATION
        if not isinstance(theme, ClTheme):
            raise ClError(
                error="Argument Error: <<theme>> must be an instance of 'calet_theme.ClTheme' class."
            )
        if not isinstance(defined, bool):
            raise ClError(
                error="Argument Error: <<defined>> must be boolean."
            )
        if not isinstance(lateral, bool):
            raise ClError(
                error="Argument Error: <<lateral>> must be boolean."
            )
        if not isinstance(expand, (bool,int)):
            raise ClError(
                error="Argument Error: <<expand>> must be boolean or integer."
            )
        if not isinstance(actions, list):
            raise ClError(
                error="Argument Error: <<actions>> must be a list."
            )
        else:
            for i in range(len(actions)):
                if not isinstance(actions[i], list):
                    raise ClError(
                        error=f"Argument Error: <<actions[{i}]>> is not a list. <<actions>> must be a list of lists."
                    )
                else:
                    for j in range(len(actions[i])):
                        if not isinstance(actions[i][j], (ClTextButton, ClButton, ClAcceptButton, ClCancelButton, ClSelectButton, ClModeButton, ClIconButton, ClMenuButton, ClCheck, ClRadio, ClSwitch)):
                            raise ClError(
                                error=f"""Argument Error: <<actions[{i}][{j}]>> must be an instance of 
                                    'calet_button.ClTextButton', 'calet_button.ClButton', 'calet_button.ClAcceptButton', 
                                    'calet_button.ClCancelButton', 'calet_button.ClSelectButton', 
                                    'calet_button.ClModeButton', 'calet_button.ClIconButton', 'calet_button.ClMenuButton', 
                                    'calet_button.ClCheck', 'calet_button.ClRadio', 'calet_button.ClSwitch' class."""
                            )
        # INITIALIZATION
        super().__init__()
        self.theme = theme
        self.actions = actions
        self.lateral = lateral
        self.defined = defined
        self.expand = expand

    def build(self):

        # SECTION
        # - creating the section
        self.section = ft.Row(
            spacing=0,
            controls=[]
        ) if not self.lateral else ft.Column(
            spacing=0,
            controls=[]
        )
        # - adding the section content
        for actions_group in self.actions:
            for action in actions_group:
                action.expand = 1
            self.section.controls.append(
                ft.Column(
                    spacing=5,
                    controls=actions_group
                ) if not self.lateral else ft.Row(
                    spacing=5,
                    controls=actions_group
                )
            )

        return ft.Container(
            bgcolor=self.theme.transparent_05 if self.defined else None,
            alignment=ft.alignment.center,
            border_radius=5,
            content=self.section
        )

# app Menu bar (ok)
class ClMenuBar(ft.UserControl):
    """Represents an app menu bar to be used in Flet apps directly or combined with a ```calet_bar.ClNavBar``` or
    ```calet_bar.ClLateralNavBar```.
    """
    def __init__(self, theme:ClTheme, sections:list[ClMenuSection], lateral:bool=False, bar_size:int=100, 
                 defined:bool=False, expand:bool|int=False, transparent:bool=False, with_blur:bool=False,
                 right_actions:list[ClTextButton|ClIconButton|ClModeButton|ClSwitch]=[]):
        """Use this properties to personalize the menu:\n
        ---
        - theme: is an instance of ```calet_theme.ClTheme``` with the colors set to paint the bar.
        - sections: is a list of ```calet_bar.MenuSection``` objects where each object will be displayed as a different section of the menu.
        - lateral: is a flag saying if the menu must be displayed verticaly or horizontaly.
        - bar_size: is the custom size of the bar. If ```expand``` is not False this property will be ignored.
        - defined: is a flag saying if the entire bar must be displayed as a defined visual block with margin from the sides of his container.
        - expand: is the responsive expansion of the menu bar in his container. See ```expand``` Flet property for more information.
        - transparent: is a flag saying if the menu must be displayed transparent or colored.
        - with_blur: is a flag saying if the menu must be displayed with blur effect or not.
        - right_actions: is a list of ```calet_button.ClTextButton```, ```calet_button.ClIconButton```, ```calet_button.ClModeButton```, ```calet_button.ClSwitch``` objects to be displayed in the right side of the menu bar.
        """
        # VALIDATION
        if not isinstance(theme, ClTheme):
            raise ClError(
                error="Argument Error: <<theme>> must be an instance of 'calet_theme.ClTheme' class."
            )
        if not isinstance(lateral, bool):
            raise ClError(
                error="Argument Error: <<lateral>> must be boolean."
            )
        if not isinstance(bar_size, int):
            raise ClError(
                error="Argument Error: <<bar_size>> must be integer."
            )
        if not isinstance(defined, bool):
            raise ClError(
                error="Argument Error: <<defined>> must be boolean."
            )
        if not isinstance(expand, (bool, int)):
            raise ClError(
                error="Argument Error: <<expand>> must be boolean or integer."
            )
        if not isinstance(transparent, bool):
            raise ClError(
                error="Argument Error: <<transparent>> must be boolean."
            )
        if not isinstance(with_blur, bool):
            raise ClError(
                error="Argument Error: <<with_blur>> must be boolean."
            )
        if not isinstance(sections, list):
            raise ClError(
                error="Argument Error: <<actions>> must be a list."
            )
        else:
            for i in range(len(sections)):
                if not isinstance(sections[i], ClMenuSection):
                    raise ClError(
                        error=f"Argument Error: <<sections[{i}]>> must be an instance of 'calet_bar.MenuSection' class."
                    )
                sections[i].lateral = lateral
        for i in range(len(right_actions)):
            if not isinstance(right_actions[i], (ClTextButton, ClIconButton, ClModeButton, ClSwitch)):
                raise ClError(
                    error=f"Argument Error: <<right_actions[{i}]>> must be an instance of 'calet_button.ClTextButton', 'calet_button.ClIconButton','calet_button.ClModeButton' or 'calet_button.ClSwitch' class."
                )
        # INITIAlIZATION
        super().__init__()
        self.theme = theme
        self.sections = sections
        self.lateral = lateral
        self.bar_size = bar_size
        self.defined = defined
        self.transparent = transparent
        self.with_blur = with_blur
        self.expand = expand
        self.right_actions = right_actions

    def build(self):

        # MENU
        # - creating the menu
        self.menu = ft.Row(
            expand=True,
            spacing=0,
            scroll=ft.ScrollMode.ADAPTIVE,
            controls=[]
        ) if not self.lateral else ft.Column(
            expand=True,
            spacing=5,
            scroll=ft.ScrollMode.ADAPTIVE,
            controls=[]
        )
        # - adding the menu content
        for section in self.sections:
            if section != self.sections[-1]:
                self.menu.controls.extend(
                    [section, ft.VerticalDivider(thickness=1, color=self.theme.divider) if not self.lateral else ft.Divider(thickness=1, color=self.theme.divider)]
                )
            else:
                self.menu.controls.append(section)
        
        # RIGHT ITEMS
        # - right items
        if self.right_actions:
            self.right_items = ft.Row(
                spacing=5,
                alignment=ft.MainAxisAlignment.END,
                controls=self.right_actions
            )

        # BAR
        self.bar = ft.Container(
            height=self.bar_size if not self.expand and not self.lateral else None,
            width=self.bar_size if not self.expand and self.lateral else None,
            bgcolor=self.theme.background_two if not self.transparent else self.theme.transparent,
            blur=5 if self.with_blur else None,
            border=ft.border.only(bottom=ft.BorderSide(1, self.theme.divider)) if not self.lateral else ft.border.only(right=ft.BorderSide(1, self.theme.divider)),
            alignment=ft.alignment.center_left if not self.lateral else ft.alignment.top_center,
            padding=5,
            animate=ft.Animation(100, ft.AnimationCurve.EASE_OUT),
            content=ft.Row(
                spacing=0,
                controls=[
                    self.menu,
                    self.right_items
                ]
            ) if self.right_actions else self.menu
        )
        if self.defined:
            self.bar.border = None
            self.bar.border_radius = 10
            self.bar.margin = ft.margin.only(
                left=5, right=5
            ) if not self.lateral else ft.margin.only(
                top=5, bottom=5
            )

        return self.bar

    def upd(self, bar_size:int=None):

        if bar_size is not None and not self.expand:
            self.bar_size = bar_size
            if self.lateral:
                self.bar.width = bar_size
            else:
                self.bar.height = bar_size
        self.update()

