from PIL import Image, ImageDraw
import copy

class SmImageController:

    def __init__(self, name:str, path:str):
        """Crea una instancia de la clase de control encargada de manejar el procesamiento de
        toda la información correspondiente a una imagen.
        ---
        - img: Es la imagen cuya información será procesada por este objeto controlador.
        - name: Es el nombre del archivo que contiene la imagen.
        - path: Es la dirección original del archivo que contiene la imagen.
        """

        self.uploads_dir = "assets/uploads"                 # Dirección de subida por defecto de las imagenes exportadas
        self.img = Image.open(path)                                     # Imagen controlada
        self.name = name[0:name.rfind(".")]                 # Nombre de la imagen sin extensión
        self.extension = name[name.rfind("."):]             # Extensión de la imagen
        self.path = path                                    # Dirección original del archivo de la imagen
        self.proportion = self.img.width/self.img.height              # Proporción de ancho por alto de la imagen original
        self.draw = ImageDraw.Draw(im=self.img, mode="RGB") # Objeto de dibujo sobre la imagen
        self.draws = [copy.deepcopy(self.img)]              # Lista de ediciones hechas sobre "img". La imagen original se conserva en la primera posición de la lista siempre.
        self.mask_imgs = []                                 # Lista de imagenes con mascaras (polígonos) para conteo de píxeles de granos.
        self.report_results = []
        self.histogram = None                               # Histograma de la imagen original.
        self.gui_width = 0                                  # ancho ocupado por la imagen en GUI
        self.gui_height = 0                                 # altura ocupada por la imagen en GUI
        self.gui_left_margin = 0                            # margen entre el borde izquierdo de la imagen y el borde izquierdo del canvas
        self.gui_top_margin = 0                             # margen entre el borde superior de la imagen y el borde superior del canvas

    def paint_circle(self, x_canvas:float, y_canvas:float, radius:float, color:str, thickness:int):
        """Dibuja un círculo sobre la imagen usando los valores dados.\n
        ---
        - x_canvas: Es la coordenada x del punto central del círculo en el canvas.
        - y_canvas: Es la coordenada y del punto central del círculo en el canvas.
        - radius: Es el radio del círculo.
        - color: Es el color del círculo.
        - thickness: Es el grosor del trazo usado para dibujar el círculo.
        """

        x = self.xcanvas_to_xreal(x_canvas=x_canvas)
        y = self.ycanvas_to_yreal(y_canvas=y_canvas)
        self.draw.ellipse(
            xy=(x, y, x+radius, y+radius),
            fill=color,
            width=thickness
        )
        self.draws.append(copy.deepcopy(self.img))
    
    def paint_line(self, x1_canvas:float, y1_canvas:float, x2_canvas:float, y2_canvas:float, color:str, thickness:int):
        """Dibuja una línea sobre la imagen usando los valores dados.\n
        ---
        - x1_canvas: Es la coordenada x del punto inicial de la línea.
        - y1_canvas: Es la coordenada y del punto inicial de la línea.
        - x2_canvas: Es la coordenada x del punto final de la línea.
        - y2_canvas: Es la coordenada y del punto final de la línea.
        - color: Es el color de la línea.
        - thickness: Es el grosor del trazo usado para dibujar la línea.
        """

        x1 = self.xcanvas_to_xreal(x_canvas=x1_canvas)
        y1 = self.ycanvas_to_yreal(y_canvas=y1_canvas)
        x2 = self.xcanvas_to_xreal(x_canvas=x2_canvas)
        y2 = self.ycanvas_to_yreal(y_canvas=y2_canvas)
        self.draw.line(
            xy=[(x1,y1), (x2,y2)],
            fill=color,
            width=thickness
        )
        self.draws.append(copy.deepcopy(self.img))

    def is_edited(self):
        return len(self.draws) > 1
    
    def undo_edition(self, is_mask:bool=False):
        """Deshace la última edición realizada sobre la imagen.\n
        ---
        - is_mask: Es una bandera indicando cuando la edición que se va a deshacer es 
                    una mascara de polígono.
        """

        if len(self.draws) > 1:
            self.draws.pop()
            self.img = copy.deepcopy(self.draws[-1])
            if is_mask:
                self.mask_imgs.pop()
    
    def clean_editions(self):
        """Elimina todas las ediciones realizadas sobre la imagen.\n
        """

        del self.draws[1:]
        self.mask_imgs.clear()
        self.img = copy.deepcopy(self.draws[0])
    
    def save_img(self, saving_path:str=None):
        """Guarda la imagen con los cambios realizados en la dirección dada. Si es None
        se guarda en la dirección por defecto.
        """

        if saving_path is not None:
            self.img.save(
                fp=saving_path if saving_path.endswith((".png",".PNG")) else f"{saving_path}.png",
                format="PNG"
            )
        else:
            self.img.save(
                fp=f"{self.uploads_dir}/{self.name}",
                format="PNG"
            )

    def xcanvas_to_xreal(self, x_canvas:str):
        """Convierte la coordenada x de un píxel del canvas a la coordenada x del píxel equivalente
        en la imagen.
        ---
        - x_canvas: Es la coordenada x del punto en el canvas.
        """

        x_gui = x_canvas - self.gui_left_margin # Quitarle el espacio del canvas que sobra a la izquierda
        w_convertion = self.img.width / self.gui_width # Razon de conversion entre ancho en gui y ancho real
        return x_gui * w_convertion

    def ycanvas_to_yreal(self, y_canvas:str):
        """Convierte la coordenada y de un píxel del canvas a la coordenada y del píxel equivalente
        en la imagen.
        ---
        - y_canvas: Es la coordenada y del punto en el canvas.
        """

        y_gui = y_canvas - self.gui_top_margin # Quitar el espacio del canvas que sobra por arriba
        h_convertion = self.img.height / self.gui_height # Razon de conversion entre altura en gui y altura real
        return y_gui * h_convertion

class SmImagesViewerController:

    def __init__(self, last_img):
        self.showed_img = 0
        self.last_img = last_img
    
    def has_previus(self):
        return True if self.showed_img > 0 else False

    def has_next(self):
        return True if self.showed_img < self.last_img else False
    
    def has_images(self):
        return self.has_next() or self.has_previus()

    def go_previus(self):
        self.showed_img = max(0, self.showed_img-1)
    
    def go_next(self):
        self.showed_img = min(self.showed_img+1, self.last_img)
    
    def go_new(self, new_last_img):
        self.showed_img += self.last_img + 1
        self.last_img = new_last_img
    
    def remove_showed(self):
        if self.showed_img == self.last_img:
            self.showed_img -= 1
        self.last_img -= 1
    
    def remove_all(self):
        self.last_img = -1
        self.showed_img = -1

class SmImagesEditorController:
    
    def __init__(self):

        self.pencil = {
            "x": 0, # x de la posicion inicial del pincel
            "y": 0, # y de la posicion inicial del pincel
            "last_x": 0, # x de la posicion final del pincel
            "last_y": 0, # y de la posicion final del pincel
            "left_limit": 0, # pixel limite del canvas hasta el que puede pintar el pincel hacia la izquierda
            "top_limit": 0, # pixel limite del canvas hasta el que puede pintar el pincel hacia arriba
            "right_limit": 0, # pixel limite del canvas hasta el que puede pintar el pincel hacia la derecha
            "bottom_limit": 0, # pixel limite del canvas hasta el que puede pintar el pincel hacia abajo
            "color": 'red', # color del pincel
            "thickness": 2
        }
        self.interval = 0 # contador de sincronización para el intervalo de dibujo
        self.drawable = False # bandera de sincronizacion para posibilidad de dibujo
        self.mode = "draw"  # bandera de modo de dibujo. Puede ser 'draw' para dibujo libre, 'crop' para recortar o 'mask' para dibujar mascara
        self.in_edition_image:SmImageController = None # referencia al controlador de la imagen en edicion

    def upd_pencil_coords(self, x:int=None, y:int=None, last_x:int=None, last_y:int=None):
        if x is not None:
            self.pencil["x"] = x
        if y is not None:
            self.pencil["y"] = y
        if last_x is not None:
            self.pencil["last_x"] = last_x
        if last_y is not None:
            self.pencil["last_y"] = last_y

    def upd_pencil_limits(self, canvas_w:int, canvas_h:int):
        img_size = self.in_edition_image.img.size
        prp = img_size[0]/img_size[1]               # razon de cambio entre altura/ancho de la foto original
        w_difference = canvas_w - img_size[0]       # diferencia entre ancho del canvas y ancho de la foto
        h_difference = canvas_h - img_size[1]       # diferencia entre altura del canvas y altura de la foto
        # si la diferencia de ancho < diferencia de altura, la img se topa antes con los bordes horizontales del canvas
        if w_difference < h_difference:             
            self.pencil["left_limit"] = 0
            self.pencil["right_limit"] = canvas_w
            self.pencil["top_limit"] = (canvas_h - (canvas_w/prp))/2
            self.pencil["bottom_limit"] = canvas_h - self.pencil["top_limit"]
        # si diferencia de ancho > diferencia de altura, la img se topa antes con los bordes verticales
        elif w_difference > h_difference: 
            self.pencil["left_limit"] = (canvas_w - (canvas_h*prp))/2
            self.pencil["right_limit"] = canvas_w - self.pencil["left_limit"]
            self.pencil["top_limit"] = 0
            self.pencil["bottom_limit"] = canvas_h
        # si la img es cuadrada, se topa antes con los bordes verticales, pero ambas dimensiones son iguales a la altura del canvas
        else: 
            self.pencil["left_limit"] = (canvas_w - canvas_h)/2
            self.pencil["right_limit"] = canvas_w - self.pencil["left_limit"]
            self.pencil["top_limit"] = 0
            self.pencil["bottom_limit"] = canvas_h
        self.in_edition_image.gui_width = self.pencil["right_limit"] - self.pencil["left_limit"]
        self.in_edition_image.gui_height = self.pencil["bottom_limit"] - self.pencil["top_limit"]
        self.in_edition_image.gui_left_margin = self.pencil["left_limit"]
        self.in_edition_image.gui_top_margin = self.pencil["top_limit"]

    def upd_pencil_stroke(self, color:str=None, thickness:int=None):
        if color is not None:
            self.pencil["color"] = color
        if thickness is not None:
            self.pencil["thickness"] = thickness

    def upd_edition_mode(self, mode:int=None, interval:int=None, drawable:bool=None):
        if mode is not None:
            self.mode = mode
        if interval is not None:
            self.interval = 0 if interval == 0 else self.interval+1
        if drawable is not None:
            self.drawable = drawable

    def can_start_draw(self, cursor_x:int, cursor_y:int):
        if self.mode != "draw":
            return False
        if not self.pencil["left_limit"] <= cursor_x <= self.pencil["right_limit"]:
            return False
        if not self.pencil["top_limit"] <= cursor_y <= self.pencil["bottom_limit"]:
            return False
        return True

    def can_update_draw(self, cursor_x:int, cursor_y:int):
        if not self.mode == "draw":
            return False
        if not self.drawable:
            return False
        if not self.pencil["left_limit"] <= cursor_x <= self.pencil["right_limit"]:
            return False
        if not self.pencil["top_limit"] <= cursor_y <= self.pencil["bottom_limit"]:
            return False
        self.interval += 1
        if not self.interval == 3:
            return False
        return True

    def can_end_draw(self):
        return self.mode == "draw"