import flet as ft
from calet.calet_button import *
from sumet_control import SmImageController, SmImagesViewerController, SmImagesEditorController

class SmImage(ft.UserControl):
    
    def __init__(self, name:str, path:str):
        """Crea la presentación visual de una imagen en la app.\n
        ---
        - name: Es el nombre del archivo con la imagen.
        - path: Es la direccion del archivo original con la imagen.
        """

        super().__init__()
        # DATOS DE LA IMAGEN
        self.expand = True                                  # NO CAMBIAR (garantiza el correcto posicionamiento de la foto
                                                            # en el que sea su contenedor)
        self.editions = ft.Stack(controls=[])               # lista de canvas superpuestos sobre la imagen para ediciones en GUI
        self.controller = SmImageController(                # Objeto controlador de esta imagen
            name=name,
            path=path
        )

    def build(self):

        # IMAGEN
        self.image = ft.Container(
            image_src=self.controller.path,
            image_fit=ft.ImageFit.CONTAIN,
            content=self.editions
        )

        return self.image
    
    # METODOS DE ACCION
    def add_canvas(self, cv:ft.canvas.Canvas):
        """Agrega un nuevo canvas al stack editions\n
        ---
        - cv: Es el nuevo canvas para la siguiente edicion.
        """

        self.editions.controls.append(cv)
        self.update()
    
    def upd_canvas(self, mode:str):
        """Actualiza el modo de edicion que representa el ultimo canvas del stack de ediciones.\n
        ---
        - mode: Es el nuevo modo de edicion del ultimo canvas
        """

        if self.editions.controls[-1].data == "crop":
            self.editions.controls[-1].shapes.clear()
        if self.editions.controls[-1].data == "mask":
            if len(self.editions.controls[-1].shapes) > 0:
                if not isinstance(self.editions.controls[-1].shapes[-1], ft.canvas.Path):
                    self.editions.controls[-1].shapes.clear()
        self.editions.controls[-1].data = mode
        self.update()
    
    def edit(self, edition):
        """Agrega una nueva edicion al ultimo canvas del stack editions.\n
        ---
        - edition: Es la nueva edicion que se hara a la foto.
        """

        # editando la foto en frontend
        if self.editions.controls[-1].data == "draw":
            self.editions.controls[-1].shapes.append(edition)
        elif self.editions.controls[-1].data == "crop":
            self.editions.controls[-1].shapes.append(edition)
        elif self.editions.controls[-1].data == "mask":
            if isinstance(edition, ft.canvas.Path):
                self.editions.controls[-1].shapes.clear()
            self.editions.controls[-1].shapes.append(edition)
        self.update()
        # procesando la edicion en backend
        if self.editions.controls[-1].data == "draw":
            if isinstance(edition, ft.canvas.Circle):
                self.controller.paint_circle(
                    x_canvas=edition.x,
                    y_canvas=edition.y,
                    radius=edition.radius,
                    color=edition.paint.color,
                    thickness=edition.paint.stroke_width
                )
            elif isinstance(edition, ft.canvas.Line):
                self.controller.paint_line(
                    x1_canvas=edition.x1,
                    y1_canvas=edition.y1,
                    x2_canvas=edition.x2,
                    y2_canvas=edition.y2,
                    color=edition.paint.color,
                    thickness=edition.paint.stroke_width
                )
        elif self.editions.controls[-1].data == "mask":
            if isinstance(edition, ft.canvas.Path):
                points_canvas = [(elem.x,elem.y) for elem in edition.elements]
                self.controller.paint_mask(
                    points_canvas=points_canvas,
                    color=edition.data,
                    tag=self.tags[-1].tag_name
                )

    def upd_edition(self, crop_width:float=None, crop_height:float=None, x1:float=None, 
                    y1:float=None, color:str=None):
        """Actualiza los valores de todas las propiedades pasadas por parametro de la ultima edicion
        realizada sobre la imagen.
        ---
        - crop_width: Es el nuevo ancho del rectangulo de recorte.
        - crop_height: Es el nuevo alto del rectangulo de recorte.
        - x1: Es la nueva x del punto inicial de la ultima edicion (o final si es una linea de delimitacion de area). 
        - y1: Es la nueva y del punto inicial de la ultima edicion (o final si es una linea de delimitacion de area).
        - color: Es el nuevo color de la ultima edicion.
        """

        if self.editions.controls[-1].data == "crop":
            self.editions.controls[-1].shapes[-1].width = crop_width
            self.editions.controls[-1].shapes[-1].height = crop_height
            self.editions.controls[-1].shapes[-1].update()
        elif self.editions.controls[-1].data == "mask":
            if color is not None:
                self.editions.controls[-1].shapes[-1].paint.color = color
                self.editions.controls[-1].shapes[-1].update()
            elif x1 is not None and y1 is not None:
                self.editions.controls[-1].shapes[-1].x2 = x1
                self.editions.controls[-1].shapes[-1].y2 = y1
                self.editions.controls[-1].shapes[-1].update()
    
    def undo(self):
        """Deshace la ultima edicion hecha sobre los canvas de la imagen.\n
        """

        # caso 1: tiene mas de 1 canvas agregado
        if len(self.editions.controls) > 1:
            current_mode = self.editions.controls[-1].data
            # caso 1.1: el ultimo canvas esta vacio (canvas que nunca se uso => descartarlo)
            if len(self.editions.controls[-1].shapes) == 0:
                self.editions.controls.pop()
            # caso 1.2: el ultimo canvas no esta vacio (canvas que se uso => limpiarlo)
            if isinstance(self.editions.controls[-1].shapes[-1], ft.canvas.Path):
                self.tags.pop()
                self.controller.undo_edition(is_mask=True)
            elif self.editions.controls[-1].data != "crop":
                self.controller.undo_edition()
            self.editions.controls[-1].shapes.clear()
            self.editions.controls[-1].data = current_mode
        # caso 2: tiene un solo canvas agregado
        else:
            self.editions.controls[-1].shapes.clear()
            self.controller.undo_edition()
        self.update()
    
    def clean(self):
        """Borra todas las ediciones hechas a la imagen.\n
        """
        self.controller.clean_editions()
        self.editions.controls.clear()
        self.tags.clear()
        self.update()

    def done(self):
        """Elimina el ultimo canvas superpuesto en la imagen si no tiene ediciones.\n
        """
        if not self.editions.controls[-1].shapes:
            self.editions.controls.pop()
        elif self.editions.controls[-1].data == "mask":
            self.editions.controls.pop()
        self.update()

class SmImagesViewer(ft.UserControl):

    def __init__(self, theme:ClTheme, images:list[SmImage], upd_app):
        super().__init__()
        self.theme = theme
        self.images = images
        self.upd_app = upd_app
        self.controller = SmImagesViewerController(last_img=len(images)-1)

    def build(self):

        # VISOR DE FOTOS
        # - foto actual
        self.actual_image = ft.Container(
            expand=True,
            alignment=ft.alignment.center,
            content=self.images[self.controller.showed_img]
        )
        # - boton 'ir a la foto anterior'
        self.b_previus_image = ClTextButton(
            theme=self.theme,
            icon=ft.icons.NAVIGATE_BEFORE,
            content_size=20,
            rounded=False,
            action=self.b_previus_image_clicked
        )
        self.b_previus_image.visible = self.controller.has_previus()
        # - boton 'ir a la foto siguiente'
        self.b_next_image = ClTextButton(
            theme=self.theme,
            icon=ft.icons.NAVIGATE_NEXT,
            content_size=20,
            rounded=False,
            action=self.b_next_image_clicked
        )
        self.b_next_image.visible = self.controller.has_next()
        # - visor
        self.viewer = ft.Container(
            expand=True,
            alignment=ft.alignment.center,
            content=ft.Row(
                spacing=0,
                alignment=ft.MainAxisAlignment.CENTER,
                controls=[
                    ft.Column(
                        expand=1,
                        spacing=0,
                        alignment=ft.MainAxisAlignment.CENTER,
                        controls=[
                            ft.Row(
                                alignment=ft.MainAxisAlignment.CENTER,
                                controls=[self.b_previus_image]
                            )
                        ]
                    ),
                    ft.Column(
                        expand=6,
                        spacing=0,
                        alignment=ft.MainAxisAlignment.CENTER,
                        controls=[self.actual_image]
                    ),
                    ft.Column(
                        expand=1,
                        spacing=0,
                        alignment=ft.MainAxisAlignment.CENTER,
                        controls=[
                            ft.Row(
                                alignment=ft.MainAxisAlignment.CENTER,
                                controls=[self.b_next_image]
                            )
                        ]
                    )
                ]
            )
        )

        return self.viewer
    
    # METODOS MANEJADORES DE EVENTOS
    def b_previus_image_clicked(self, e:ft.TapEvent):
        self.controller.go_previus()
        self.actual_image.content = self.images[self.controller.showed_img]
        self.upd_app(title=f"SuMetApp - {self.images[self.controller.showed_img].controller.name}")
        self.b_previus_image.visible = self.controller.has_previus()
        self.b_next_image.visible = True
        self.update()
    
    def b_next_image_clicked(self, e:ft.TapEvent):
        self.controller.go_next()
        self.actual_image.content = self.images[self.controller.showed_img]
        self.upd_app(title=f"SuMetApp - {self.images[self.controller.showed_img].controller.name}")
        self.b_previus_image.visible = True
        self.b_next_image.visible = self.controller.has_next()
        self.update()

    # METODOS DE ACCION
    def showed_image(self):
        return self.images[self.controller.showed_img]

    def add_images(self, images:list[SmImage]):

        self.images = self.images + images
        self.actual_image.content = images[0]
        self.upd_app(title=f"SuMetApp - {images[0].controller.name}")
        self.controller.go_new(new_last_img=len(self.images)-1)
        self.b_previus_image.visible = self.controller.has_previus()
        self.b_next_image.visible = self.controller.has_next()
        self.update()
    
    def discard_images(self, all:bool=False):
        """Descarta determinadas imagenes cargadas.\n
        ---
        - all: Es una bandera indicando si deben ser descartadas todas las imagenes o solo
                la que se esta visualizando.
        """
        if all:
            self.images.clear()
            self.controller.remove_all()
        else:
            self.images.pop(self.controller.showed_img)
            self.controller.remove_showed()
            if self.controller.has_images():
                self.actual_image.content = self.images[self.controller.showed_img]
                self.b_previus_image.visible = self.controller.has_previus()
                self.b_next_image.visible = self.controller.has_next()
        self.update()

    def lock_carrousel(self):
        self.b_previus_image.visible = False
        self.b_next_image.visible = False
        self.update()
    
    def unlock_carrousel(self):
        self.b_previus_image.visible = self.controller.has_previus()
        self.b_next_image.visible = self.controller.has_next()
        self.update()

class SmEditorToolsBar(ft.UserControl):
    
    def __init__(self, theme:ClTheme, items:list=[], hidden_x=-1.1, showed_x=0, title:str=None, expand=False):
        """Crea una barra de acciones lateral. Las barras de acciones laterales se usan para
        mostrar acciones especificas de cada estado de la app, como por ejemplo las de 
        edicion.
        ---
        - theme: Es el tema actual de la app.
        - items: Es la lista de elementos que se mostraran en la barra de acciones.
        - expand: Es una bandera indicando cuando la barra lateral debe expandirse o no.
        """

        super().__init__()
        self.theme = theme
        self.items = items
        self.hidden_x = hidden_x
        self.showed_x = showed_x
        self.title = title
        self.offset = ft.Offset(hidden_x,0)
        self.animate_offset = 100
        self.expand = expand
    
    def build(self):

        # BARRA DE ACCIONES
        # - titulo de la barra
        if self.title is not None:
            self.bar_title = ft.Row(
                alignment=ft.MainAxisAlignment.START,
                controls=[
                    ft.Container(
                        alignment=ft.alignment.center,
                        margin=ft.margin.only(left=10),
                        content=ft.Text(
                            value=self.title,
                            color=self.theme.font_one,
                            size=11,
                            text_align=ft.TextAlign.CENTER
                        )
                    )
                ]
            )
        # - barra
        self.bar = ft.Container(
            padding=5,
            border_radius=10,
            bgcolor=self.theme.background_two,
            alignment=ft.alignment.center,
            margin=ft.margin.only(left=10),
            shadow=ft.BoxShadow(
                spread_radius=0,
                blur_radius=2,
                color="black",
                blur_style=ft.ShadowBlurStyle.OUTER
            ),
            content=ft.Column(
                spacing=5,
                controls=self.items
            )
        )

        return ft.Column(
            spacing=10,
            alignment=ft.MainAxisAlignment.CENTER,
            controls=[
                ft.Row(controls=[self.bar_title]),
                ft.Row(controls=[self.bar])
            ] if self.title is not None else [
                ft.Row(controls=[self.bar])
            ]
        )
    
    # METODOS DE ACCION
    def show_bar(self):
        self.offset.x = self.showed_x
        self.update()

    def hide_bar(self):
        self.offset.x = self.hidden_x
        self.update()

    def showed(self):
        return self.offset.x == self.showed_x

class SmImagesEditor(ft.UserControl):
    
    def __init__(self, theme:ClTheme):
        """Crea un editor de imagenes para la app. Este se encargara de gestionar todas
        las acciones y eventos relacionados con la edicion y visualizacion de imagenes
        cargadas en la app.
        ---
        - theme: Es el tema actual de la app.
        - notify: Es una referencia al metodo de notificar definido en la clase
                    SmContentBody.
        """

        super().__init__()
        self.theme = theme
        self.in_edition_image = None
        self.controller = SmImagesEditorController()

    def build(self):

        # BARRA FLOTANTE DE CONFIGURACIÓN DEL PINCEL
        # - boton 'paleta de colores del pincel'
        self.b_palette = ClNavButton(
            theme=self.theme,
            label="Colores",
            icon=ft.icons.COLOR_LENS_OUTLINED,
            selected_icon=ft.icons.COLOR_LENS,
            content_size=16,
            width=60,
            rounded=False,
            action=self.b_palette_clicked
        )
        # - boton 'grosor del trazo del pincel'
        self.b_thickness = ClNavButton(
            theme=self.theme,
            label="Grosor del trazo",
            icon=ft.icons.LINE_WEIGHT_OUTLINED,
            selected_icon=ft.icons.LINE_WEIGHT,
            content_size=16,
            width=60,
            rounded=False,
            action=self.b_thickness_clicked
        )
        # - barra de configuracion
        self.config_toolsbar = SmEditorToolsBar(
            theme=self.theme,
            title="Configuración",
            items=[
                self.b_palette,
                self.b_thickness,
            ]
        )

        # BARRA FLOTANTE DE MODOS DE PINCEL
        # - boton 'modo dibujo libre'
        self.b_draw = ClNavButton(
            theme=self.theme,
            label="Dibujar",
            icon=ft.icons.DRAW_OUTLINED,
            selected_icon=ft.icons.DRAW,
            content_size=16,
            width=60,
            rounded=False,
            action=self.b_draw_clicked
        )
        # - barra de modos
        self.modes_toolsbar = SmEditorToolsBar(
            theme=self.theme,
            title="Herramientas",
            items=[
                self.b_draw,
                self.b_crop,
                self.b_mask,
            ],
        )

        # BARRA FLOTANTE CON LA PALETA DE COLORES
        # - blanco
        self.b_white = ClColorButton(
            theme=self.theme,
            color=ft.colors.WHITE,
            width=40,
            action=self.b_color_clicked
        )
        # - negro
        self.b_black = ClColorButton(
            theme=self.theme,
            color=ft.colors.BLACK,
            width=40,
            action=self.b_color_clicked
        )
        # - marron
        self.b_brown = ClColorButton(
            theme=self.theme,
            color=ft.colors.BROWN,
            width=40,
            action=self.b_color_clicked
        )
        # - rojo
        self.b_red = ClColorButton(
            theme=self.theme,
            color=ft.colors.RED,
            width=40,
            selected=True,
            action=self.b_color_clicked
        )
        # - naranja
        self.b_orange = ClColorButton(
            theme=self.theme,
            color=ft.colors.ORANGE,
            width=40,
            action=self.b_color_clicked
        )
        # - amarillo
        self.b_yellow = ClColorButton(
            theme=self.theme,
            color=ft.colors.YELLOW,
            width=40,
            action=self.b_color_clicked
        )
        # - verde
        self.b_green = ClColorButton(
            theme=self.theme,
            color=ft.colors.GREEN,
            width=40,
            action=self.b_color_clicked
        )
        # - cian
        self.b_cyan = ClColorButton(
            theme=self.theme,
            color=ft.colors.CYAN,
            width=40,
            action=self.b_color_clicked
        )
        # - azul
        self.b_blue = ClColorButton(
            theme=self.theme,
            color=ft.colors.BLUE,
            width=40,
            action=self.b_color_clicked
        )
        # - indigo
        self.b_indigo = ClColorButton(
            theme=self.theme,
            color=ft.colors.INDIGO,
            width=40,
            action=self.b_color_clicked
        )
        # - violeta
        self.b_purple = ClColorButton(
            theme=self.theme,
            color=ft.colors.PURPLE,
            width=40,
            action=self.b_color_clicked
        )
        # - paleta de colores
        self.palette_toolsbar = SmEditorToolsBar(
            theme=self.theme,
            showed_x=1.4,
            items=[
                self.b_white,
                self.b_black,
                self.b_brown,
                self.b_red,
                self.b_orange,
                self.b_yellow,
                self.b_green,
                self.b_cyan,
                self.b_blue,
                self.b_indigo,
                self.b_purple,
            ]
        )

        # BARRA FLOTANTE CON OPCIONES DE GROSOR DEL PINCEL
        # - boton 'aumentar grosor'
        self.b_add_thickness = ClTextButton(
            theme=self.theme,
            icon=ft.icons.ADD,
            width=40,
            height=40,
            action=self.b_add_thickness_clicked
        )
        # - boton 'disminuir grosor'
        self.b_remove_thickness = ClTextButton(
            theme=self.theme,
            icon=ft.icons.REMOVE,
            width=40,
            height=40,
            action=self.b_remove_thickness_clicked
        )
        # - contador de grosor
        self.thickness_counter = ft.Container(
            alignment=ft.alignment.center,
            bgcolor=self.theme.secondary_block,
            padding=10,
            width=40,
            height=40,
            border_radius=5,
            content=ft.Text(
                value=self.controller.pencil["thickness"],
                color=self.theme.primary,
                text_align=ft.TextAlign.CENTER,
                weight=ft.FontWeight.BOLD
            )
        )
        # - barra de botones
        self.thickness_toolsbar = SmEditorToolsBar(
            theme=self.theme,
            showed_x=1.4,
            items=[
                self.b_add_thickness,
                self.thickness_counter,
                self.b_remove_thickness
            ]
        )

        # EDITOR
        self.editor = ft.Stack(
            controls=[
                ft.Column(
                    expand=True,
                    alignment=ft.MainAxisAlignment.CENTER,
                    controls=[
                        ft.Row(expand=True),
                        ft.Row(
                            alignment=ft.MainAxisAlignment.START,
                            controls=[self.palette_toolsbar]
                        ),
                        ft.Row(expand=True)
                    ]
                ),
                ft.Column(
                    expand=True,
                    alignment=ft.MainAxisAlignment.CENTER,
                    controls=[
                        ft.Row(expand=True),
                        ft.Row(
                            alignment=ft.MainAxisAlignment.START,
                            controls=[self.thickness_toolsbar]
                        ),
                        ft.Row(expand=True)
                    ]
                ),
                ft.Column(
                    expand=True,
                    alignment=ft.MainAxisAlignment.CENTER,
                    controls=[
                        ft.Row(expand=True),
                        ft.Row(
                            alignment=ft.MainAxisAlignment.START,
                            controls=[self.config_toolsbar]
                        ),
                        ft.Row(
                            alignment=ft.MainAxisAlignment.START,
                            controls=[self.modes_toolsbar]
                        ),
                        ft.Row(expand=True)
                    ]
                ),
            ]
        )

        return self.editor
    
    # METODOS MANEJADORES DE EVENTOS
    def b_palette_clicked(self, e:ft.TapEvent):
        if self.b_palette.selected:
            self.palette_toolsbar.show_bar()
            if self.b_thickness.selected:
                self.b_thickness.upd(selected=False)
                self.thickness_toolsbar.hide_bar()
        else:
            self.palette_toolsbar.hide_bar()
        self.update()

    def b_color_clicked(self, e:ft.TapEvent):
        {
            ft.colors.WHITE: self.b_white,
            ft.colors.BLACK: self.b_black,
            ft.colors.BROWN: self.b_brown,
            ft.colors.RED: self.b_red,
            ft.colors.ORANGE: self.b_orange,
            ft.colors.YELLOW: self.b_yellow,
            ft.colors.GREEN: self.b_green,
            ft.colors.CYAN: self.b_cyan,
            ft.colors.BLUE: self.b_blue,
            ft.colors.INDIGO: self.b_indigo,
            ft.colors.PURPLE: self.b_purple,
        }[self.controller.pencil["color"]].upd(selected=False)
        self.controller.upd_pencil_stroke(color=e.control.data.color)
        self.update()

    def b_thickness_clicked(self, e:ft.TapEvent):
        if self.b_thickness.selected:
            self.thickness_toolsbar.show_bar()
            if self.b_palette.selected:
                self.b_palette.upd(selected=False)
                self.palette_toolsbar.hide_bar()
        else:
            self.thickness_toolsbar.hide_bar()
        self.update()

    def b_add_thickness_clicked(self, e:ft.TapEvent):
        thickness = min(int(self.thickness_counter.content.value)+1, 9)
        self.controller.upd_pencil_stroke(thickness=thickness)
        self.thickness_counter.content.value = thickness
        self.thickness_counter.content.update()
        self.update()

    def b_remove_thickness_clicked(self, e:ft.TapEvent):
        thickness = max(1, int(self.thickness_counter.content.value)-1)
        self.controller.upd_pencil_stroke(thickness=thickness)
        self.thickness_counter.content.value = thickness
        self.thickness_counter.content.update()
        self.update()

    def b_draw_clicked(self, e:ft.TapEvent):
        self.controller.upd_edition_mode(mode="draw", interval=0, drawable=False)
        self.in_edition_image.upd_canvas(mode="draw")
        self.b_draw.upd(selected=True)
        if self.b_crop.selected:
            self.b_crop.upd(selected=False)
            if self.crop_toolsbar.showed():
                self.crop_toolsbar.hide_bar()
        if self.b_mask.selected:
            self.b_mask.upd(selected=False)
            if self.mask_toolsbar.showed():
                self.crop_toolsbar.hide_bar()
        if self.b_palette.selected:
            self.b_palette.upd(selected=False)
            self.palette_toolsbar.hide_bar()
        if self.b_thickness.selected:
            self.b_thickness.upd(selected=False)
            self.thickness_toolsbar.hide_bar()
        self.update()
     
    def cv_resized(self, e:ft.canvas.CanvasResizeEvent):
        if self.in_edition_image is not None:
            self.controller.upd_pencil_limits(
                canvas_w=e.width,
                canvas_h=e.height
            )

    def gd_pan_started(self, e:ft.DragStartEvent):
        if self.controller.can_start_draw(cursor_x=e.local_x, cursor_y=e.local_y):
            self.in_edition_image.edit(
                ft.canvas.Circle(
                    x=e.local_x,
                    y=e.local_y,
                    radius=self.controller.pencil["thickness"]/2,
                    paint=ft.Paint(
                        color=self.controller.pencil["color"],
                        stroke_width=self.controller.pencil["thickness"],
                        stroke_join=ft.StrokeJoin.ROUND,
                        stroke_cap=ft.StrokeCap.ROUND
                    )
                )
            )
            self.controller.upd_pencil_coords(x=e.local_x, y=e.local_y)
            self.controller.upd_edition_mode(interval=1, drawable=True)
            self.update()
        elif self.controller.can_start_crop(cursor_x=e.local_x, cursor_y=e.local_y):
            if self.crop_toolsbar.showed():
                self.crop_toolsbar.hide_bar()
                self.in_edition_image.undo()
            self.in_edition_image.edit(
                ft.canvas.Circle(
                    x=e.local_x,
                    y=e.local_y,
                    radius=5,
                    paint=ft.Paint(
                        color=self.controller.pencil["color"],
                        stroke_width=self.controller.pencil["thickness"],
                        stroke_join=ft.StrokeJoin.ROUND,
                        stroke_cap=ft.StrokeCap.ROUND
                    )
                )
            )
            self.controller.upd_pencil_coords(x=e.local_x, y=e.local_y)
            self.controller.upd_edition_mode(mode="crop_started", drawable=True)
            self.update()

    def gd_pan_updated(self, e:ft.DragUpdateEvent):
        if self.controller.can_update_draw(cursor_x=e.local_x, cursor_y=e.local_y):
            self.in_edition_image.edit(
                ft.canvas.Line(
                    x1=self.controller.pencil["x"],
                    y1=self.controller.pencil["y"],
                    x2=e.local_x,
                    y2=e.local_y,
                    paint=ft.Paint(
                        color=self.controller.pencil["color"],
                        stroke_width=self.controller.pencil["thickness"],
                        stroke_join=ft.StrokeJoin.ROUND,
                        stroke_cap=ft.StrokeCap.ROUND
                    )
                )
            )
            self.controller.upd_pencil_coords(x=e.local_x, y=e.local_y)
            self.controller.upd_edition_mode(interval=0)
            self.update()
        elif self.controller.can_update_crop(cursor_x=e.local_x, cursor_y=e.local_y):
            self.in_edition_image.edit(
                ft.canvas.Rect(
                    x=self.controller.pencil["x"],
                    y=self.controller.pencil["y"],
                    width=e.local_x-self.controller.pencil["x"],
                    height=e.local_y-self.controller.pencil["y"],
                    paint=ft.Paint(
                        color=self.controller.pencil["color"],
                        stroke_width=2,
                        stroke_cap=ft.StrokeCap.ROUND,
                        stroke_join=ft.StrokeJoin.ROUND,
                        style=ft.PaintingStyle.STROKE
                    )
                )
            )
            self.controller.upd_pencil_coords(last_x=e.local_x, last_y=e.local_y)
            self.controller.upd_edition_mode(mode="on_crop", interval=0)
            self.update()
        elif self.controller.can_update_croprectangle(cursor_x=e.local_x, cursor_y=e.local_y):
            self.in_edition_image.upd_edition(
                crop_width=e.local_x-self.controller.pencil["x"],
                crop_height=e.local_y-self.controller.pencil["y"],
            )
            self.controller.upd_pencil_coords(last_x=e.local_x, last_y=e.local_y)
            self.controller.upd_edition_mode(interval=0)
            self.update()

    def gd_pan_ended(self, e:ft.DragEndEvent):
        if self.controller.can_end_draw():
            self.in_edition_image.add_canvas(
                ft.canvas.Canvas(
                    data="draw",
                    expand=True,
                    shapes=[],
                    content=ft.GestureDetector(
                        on_pan_start=self.gd_pan_started,
                        on_pan_update=self.gd_pan_updated,
                        on_pan_end=self.gd_pan_ended,
                        on_hover=self.gd_hovered,
                        on_tap=self.gd_tapped
                    )
                )
            )
            self.controller.upd_edition_mode(interval=0, drawable=False)
            self.update()
        elif self.controller.can_end_crop():
            self.in_edition_image.edit(
                ft.canvas.Circle(
                    x=self.controller.pencil["last_x"],
                    y=self.controller.pencil["last_y"],
                    radius=5,
                    paint=ft.Paint(
                        color=self.controller.pencil["color"],
                        stroke_width=self.controller.pencil["thickness"],
                        stroke_join=ft.StrokeJoin.ROUND,
                        stroke_cap=ft.StrokeCap.ROUND
                    )
                )
            )
            self.crop_toolsbar.show_bar()
            self.controller.upd_edition_mode(mode="crop", interval=0, drawable=False)
            self.update()

    def gd_hovered(self, e:ft.HoverEvent):
        if self.controller.mode in ("mask", "mask_started", "mask_updating"):
            self.controller.upd_pencil_coords(last_x=e.local_x, last_y=e.local_y)
        if self.controller.can_start_maskpath(cursor_x=e.local_x, cursor_y=e.local_y):
            self.in_edition_image.edit(
                ft.canvas.Line(
                    x1=self.controller.pencil["x"],
                    y1=self.controller.pencil["y"],
                    x2=e.local_x,
                    y2=e.local_y,
                    paint=ft.Paint(
                        color=ft.colors.with_opacity(0.5, self.controller.pencil["color"]),
                        stroke_width=2,
                        stroke_cap=ft.StrokeCap.ROUND,
                        stroke_join=ft.StrokeJoin.ROUND
                    )
                )
            )
            self.controller.upd_edition_mode(mode="mask_updating", interval=0)
            self.update()
        elif self.controller.can_update_maskpath(cursor_x=e.local_x, cursor_y=e.local_y):
            self.in_edition_image.upd_edition(
                x1=e.local_x, 
                y1=e.local_y
            )
            self.controller.upd_edition_mode(interval=0)
            self.update()
    
    def gd_tapped(self, e:ft.ControlEvent):
        if self.controller.can_start_mask() or self.controller.can_update_mask():
            self.in_edition_image.edit(
                ft.canvas.Circle(
                    x=self.controller.pencil["last_x"],
                    y=self.controller.pencil["last_y"],
                    radius=5,
                    paint=ft.Paint(
                        color=self.controller.pencil["color"],
                        stroke_width=self.controller.pencil["thickness"],
                        stroke_cap=ft.StrokeCap.ROUND,
                        stroke_join=ft.StrokeJoin.ROUND
                    )
                )
            )
            self.controller.upd_edition_mode(mode="mask_started", interval=0, drawable=True)
            self.controller.upd_pencil_coords(x=self.controller.pencil["last_x"], y=self.controller.pencil["last_y"])
            if self.controller.can_update_mask():
                self.in_edition_image.upd_edition(color=self.controller.pencil["color"])
            self.mask_toolsbar.show_bar()
            self.update()

    # METODOS DE ACCION
    def start_edition(self, image:SmImage):
        image.add_canvas(
            cv=ft.canvas.Canvas(
                data="draw",
                expand=True,
                shapes=[],
                on_resize=self.cv_resized,
                content=ft.GestureDetector(
                    on_pan_start=self.gd_pan_started,
                    on_pan_update=self.gd_pan_updated,
                    on_pan_end=self.gd_pan_ended,
                    on_hover=self.gd_hovered,
                    on_tap=self.gd_tapped
                )
            )
        )
        self.in_edition_image = image
        self.controller.in_edition_image = image.controller
        self.config_toolsbar.show_bar()
        self.modes_toolsbar.show_bar()
        self.b_draw.upd(selected=True)
        self.update()
    
    def undo_edition(self):
        self.in_edition_image.undo()
        self.controller.upd_edition_mode(interval=0, drawable=False)
        self.update()

    def finish_edition(self):
        self.controller.upd_edition_mode(mode="draw", interval=0, drawable=False)
        self.b_palette.upd(selected=False)
        self.b_thickness.upd(selected=False)
        self.b_draw.upd(selected=False)
        self.b_crop.upd(selected=False)
        self.b_mask.upd(selected=False)
        self.config_toolsbar.hide_bar()
        self.modes_toolsbar.hide_bar()
        if self.palette_toolsbar.showed():
            self.palette_toolsbar.hide_bar()
        if self.palette_toolsbar.showed():
            self.thickness_toolsbar.hide_bar()
        if self.crop_toolsbar.showed():
            self.crop_toolsbar.hide_bar()
            self.in_edition_image.undo()
        if self.mask_toolsbar.showed():
            self.mask_toolsbar.hide_bar()
        if self.tagsbar.showed():
            self.tagsbar.clean()
            self.tagsbar.hide_bar()
        self.in_edition_image = None
        self.controller.in_edition_image = None
        self.update()

    def confirm_edition(self):
        self.in_edition_image.done()
        self.finish_edition()

    def cancel_edition(self):
        self.in_edition_image.clean()
        self.finish_edition()